/*
 * PCMSolver, an API for the Polarizable Continuum Model
 * Copyright (C) 2016 Roberto Di Remigio, Luca Frediani and collaborators.
 *
 * This file is part of PCMSolver.
 *
 * PCMSolver is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * PCMSolver is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with PCMSolver.  If not, see <http://www.gnu.org/licenses/>.
 *
 * For information on the complete list of contributors to the
 * PCMSolver API, see: <http://pcmsolver.readthedocs.io/>
 */

#ifndef C_HOST_FUNCTIONS_H
#define C_HOST_FUNCTIONS_H
#include <math.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "PCMInput.h"

#if (defined(__STDC__) && (__STDC_VERSION__ < 199901L)) && !defined(__cplusplus)
typedef enum { false, true } bool;
#else /* (defined(__STDC__) || (__STDC_VERSION__ < 199901L)) &&                     \
         !defined(__cplusplus) */
#include <stdbool.h>
#endif /* (defined(__STDC__) || (__STDC_VERSION__ < 199901L)) &&                    \
          !defined(__cplusplus) */

#ifdef __GNUC__
#define UNUSED(x) UNUSED_##x __attribute__((__unused__))
#else
#define UNUSED(x) UNUSED_##x
#endif

#ifdef __GNUC__
#define UNUSED_FUNCTION(x) __attribute__((__unused__)) UNUSED_##x
#else
#define UNUSED_FUNCTION(x) UNUSED_##x
#endif

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

void host_writer(const char * message);

struct PCMInput pcmsolver_input();

/*! \brief calculates nuclear molecular electrostatic potential (MEP) at cavity
 * points
 *  \author Roberto Di Remigio
 *  \date 2015
 *  \param nr_nuclei number of atomic centers
 *  \param charges atomic charges
 *  \param coordinates coordinates of the atomic centers
 *  \param grid_size the number of cavity points
 *  \param grid the cavity points
 *  \return the nuclear MEP
 *  \warning Caller deallocats returned array
 */
double * nuclear_mep(int nr_nuclei,
                     double charges[nr_nuclei],
                     double coordinates[3 * nr_nuclei],
                     int grid_size,
                     double grid[3 * grid_size]);

/*! \brief Compares calculated and reference values within a threshold
 *  \author Roberto Di Remigio
 *  \date 2015
 *  \param calculated calculated value
 *  \param reference  reference value
 *  \param threshold  comparison threshold
 *  \return true if the calculated and reference values are compatible, false
 * otherwise
 */
bool check_unsigned_error(double calculated, double reference, double threshold);

int test_surface_functions(FILE * fp,
                           int grid_size,
                           double mep[grid_size],
                           double asc_Ag[grid_size],
                           double asc_B3g[grid_size],
                           double asc_neq_B3g[grid_size],
                           double areas[grid_size]);

double mep_reference(int i);
double asc_Ag_reference(int i);
double asc_B3g_reference(int i);
double asc_neq_B3g_reference(int i);
double areas_reference(int i);

struct PCMInput pcmsolver_input() {
  struct PCMInput host_input;

  // These parameters would be set by the host input reading
  // Length and area parameters are all assumed to be in Angstrom,
  // the module will convert to Bohr internally
  strcpy(host_input.cavity_type, "gepol");
  host_input.patch_level = 2;
  host_input.coarsity = 0.5;
  host_input.area = 0.2;
  host_input.min_distance = 0.1;
  host_input.der_order = 4;
  host_input.scaling = true;
  strcpy(host_input.radii_set, "bondi");
  strcpy(host_input.restart_name, "cavity.npz");
  host_input.min_radius = 100.0;

  strcpy(host_input.solver_type, "iefpcm");
  strcpy(host_input.solvent, "water");
  strcpy(host_input.equation_type, "secondkind");
  host_input.correction = 0.0;
  host_input.probe_radius = 1.0;

  strcpy(host_input.inside_type, "vacuum");
  host_input.outside_epsilon = 1.0;
  strcpy(host_input.outside_type, "uniformdielectric");

  return host_input;
}

double * nuclear_mep(int nr_nuclei,
                     double charges[nr_nuclei],
                     double coordinates[3 * nr_nuclei],
                     int grid_size,
                     double grid[3 * grid_size]) {
  double * mep = (double *)calloc(grid_size, sizeof(double));
  for (int i = 0; i < nr_nuclei; i++) {
    for (int j = 0; j < grid_size; j++) {
      // Column-major ordering. Offsets: col_idx * nr_rows + row_idx
      double dist = pow((coordinates[i * 3] - grid[j * 3]), 2) +
                    pow((coordinates[i * 3 + 1] - grid[j * 3 + 1]), 2) +
                    pow((coordinates[i * 3 + 2] - grid[j * 3 + 2]), 2);
      mep[j] += charges[i] / sqrt(dist);
    }
  }
  return mep;
}

bool check_unsigned_error(double calculated, double reference, double threshold) {
  double err = calculated - reference;
  return (fabs(err) <= fabs(calculated) * threshold);
}

int test_surface_functions(FILE * fp,
                           int grid_size,
                           double mep[grid_size],
                           double asc_Ag[grid_size],
                           double asc_B3g[grid_size],
                           double asc_neq_B3g[grid_size],
                           double areas[grid_size]) {
  const int MEP_ERROR = 1;
  const int ASC_Ag_ERROR = 2;
  const int ASC_B3g_ERROR = 3;
  const int ASC_NEQ_B3g_ERROR = 4;
  const int AREAS_ERROR = 5;
  for (int i = 0; i < grid_size; i++) {
    if (!check_unsigned_error(mep[i], mep_reference(i), 1.0e-07)) {
      fprintf(stderr,
              "%s\n",
              "Error in MEP, please file an issue on: "
              "https://github.com/PCMSolver/pcmsolver");
      return MEP_ERROR;
    }
    if (!check_unsigned_error(asc_Ag[i], asc_Ag_reference(i), 1.0e-07)) {
      fprintf(stderr,
              "%s\n",
              "Error in ASC Ag, please file an issue on: "
              "https://github.com/PCMSolver/pcmsolver");
      return ASC_Ag_ERROR;
    }
    if (!check_unsigned_error(asc_B3g[i], asc_B3g_reference(i), 1.0e-07)) {
      fprintf(stderr,
              "%s\n",
              "Error in ASC B3g, please file an issue on: "
              "https://github.com/PCMSolver/pcmsolver");
      return ASC_B3g_ERROR;
    }
    if (!check_unsigned_error(asc_neq_B3g[i], asc_neq_B3g_reference(i), 1.0e-07)) {
      fprintf(stderr,
              "%s\n",
              "Error in nonequilibrium ASC B3g, please file an "
              "issue on: https://github.com/PCMSolver/pcmsolver");
      return ASC_NEQ_B3g_ERROR;
    }
    if (!check_unsigned_error(areas[i], areas_reference(i), 1.0e-07)) {
      fprintf(stderr,
              "%s\n",
              "Error in finite elements areas, please file an issue "
              "on: https://github.com/PCMSolver/pcmsolver");
      return AREAS_ERROR;
    }
  }
  fprintf(fp, "%s\n", "Test on MEP: PASSED");
  fprintf(fp, "%s\n", "Test on ASC Ag: PASSED");
  fprintf(fp, "%s\n", "Test on ASC B3g: PASSED");
  fprintf(fp, "%s\n", "Test on nonequilibrium ASC in B3g symmetry: PASSED");
  fprintf(fp, "%s\n", "Test on finite element areas: PASSED");

  return 0;
}

double mep_reference(int i) {
  static double mep[576] = {
      3.372240447717, 3.395095348797, 3.375472982012, 3.407830928930, 3.389673049936,
      3.444734322778, 3.434450634681, 3.567420410364, 3.519403897623, 3.521500360692,
      3.737204637428, 3.665836385359, 3.638475909577, 3.634697522671, 3.378728206062,
      3.400511009364, 3.387155190391, 3.430448505960, 3.414568660060, 3.502974792852,
      3.485472605945, 3.646792406411, 3.600687428544, 3.596042090542, 4.043143323646,
      4.042627673887, 4.000207071572, 3.943830109299, 3.901090644538, 3.971528385050,
      3.922106005834, 3.832443093990, 3.781559803821, 3.745079273102, 3.726956468823,
      4.071479016092, 4.039193036616, 3.982347524785, 3.939905876613, 3.970318175141,
      3.926366695705, 3.883604616653, 3.829224917150, 3.786789784941, 3.137381224862,
      3.001405458723, 3.186612154619, 3.000297605619, 3.091578048072, 3.363118040231,
      3.133206449617, 3.182855203960, 3.381998093999, 3.087995448210, 2.999536488978,
      3.192470183633, 3.085172887610, 3.190455940555, 3.445493468024, 3.663096576477,
      3.766481150018, 3.271760664144, 3.368194930794, 3.540250609025, 3.824911768840,
      3.397511985055, 3.568518356734, 3.268466398614, 3.329351334592, 3.421372568835,
      3.333950027352, 3.407773545920, 3.372240447717, 3.395095348797, 3.375472982012,
      3.407830928930, 3.389673049936, 3.444734322778, 3.434450634681, 3.567420410364,
      3.519403897623, 3.521500360692, 3.737204637428, 3.665836385359, 3.638475909577,
      3.634697522671, 3.378728206062, 3.400511009364, 3.387155190391, 3.430448505960,
      3.414568660060, 3.502974792852, 3.485472605945, 3.646792406411, 3.600687428544,
      3.596042090542, 4.043143323646, 4.042627673887, 4.000207071572, 3.943830109299,
      3.901090644538, 3.971528385050, 3.922106005834, 3.832443093990, 3.781559803821,
      3.745079273102, 3.726956468823, 4.071479016092, 4.039193036616, 3.982347524785,
      3.939905876613, 3.970318175141, 3.926366695705, 3.883604616653, 3.829224917150,
      3.786789784941, 3.137381224862, 3.001405458723, 3.186612154619, 3.000297605619,
      3.091578048072, 3.363118040231, 3.133206449617, 3.182855203960, 3.381998093999,
      3.087995448210, 2.999536488978, 3.192470183633, 3.085172887610, 3.190455940555,
      3.445493468024, 3.663096576477, 3.766481150018, 3.271760664144, 3.368194930794,
      3.540250609025, 3.824911768840, 3.397511985055, 3.568518356734, 3.268466398614,
      3.329351334592, 3.421372568835, 3.333950027352, 3.407773545920, 3.372240447717,
      3.395095348797, 3.375472982012, 3.407830928930, 3.389673049936, 3.444734322778,
      3.434450634681, 3.567420410364, 3.519403897623, 3.521500360692, 3.737204637428,
      3.665836385359, 3.638475909577, 3.634697522671, 3.378728206062, 3.400511009364,
      3.387155190391, 3.430448505960, 3.414568660060, 3.502974792852, 3.485472605945,
      3.646792406411, 3.600687428544, 3.596042090542, 4.043143323646, 4.042627673887,
      4.000207071572, 3.943830109299, 3.901090644538, 3.971528385050, 3.922106005834,
      3.832443093990, 3.781559803821, 3.745079273102, 3.726956468823, 4.071479016092,
      4.039193036616, 3.982347524785, 3.939905876613, 3.970318175141, 3.926366695705,
      3.883604616653, 3.829224917150, 3.786789784941, 3.137381224862, 3.001405458723,
      3.186612154619, 3.000297605619, 3.091578048072, 3.363118040231, 3.133206449617,
      3.182855203960, 3.381998093999, 3.087995448210, 2.999536488978, 3.192470183633,
      3.085172887610, 3.190455940555, 3.445493468024, 3.663096576477, 3.766481150018,
      3.271760664144, 3.368194930794, 3.540250609025, 3.824911768840, 3.397511985055,
      3.568518356734, 3.268466398614, 3.329351334592, 3.421372568835, 3.333950027352,
      3.407773545920, 3.372240447717, 3.395095348797, 3.375472982012, 3.407830928930,
      3.389673049936, 3.444734322778, 3.434450634681, 3.567420410364, 3.519403897623,
      3.521500360692, 3.737204637428, 3.665836385359, 3.638475909577, 3.634697522671,
      3.378728206062, 3.400511009364, 3.387155190391, 3.430448505960, 3.414568660060,
      3.502974792852, 3.485472605945, 3.646792406411, 3.600687428544, 3.596042090542,
      4.043143323646, 4.042627673887, 4.000207071572, 3.943830109299, 3.901090644538,
      3.971528385050, 3.922106005834, 3.832443093990, 3.781559803821, 3.745079273102,
      3.726956468823, 4.071479016092, 4.039193036616, 3.982347524785, 3.939905876613,
      3.970318175141, 3.926366695705, 3.883604616653, 3.829224917150, 3.786789784941,
      3.137381224862, 3.001405458723, 3.186612154619, 3.000297605619, 3.091578048072,
      3.363118040231, 3.133206449617, 3.182855203960, 3.381998093999, 3.087995448210,
      2.999536488978, 3.192470183633, 3.085172887610, 3.190455940555, 3.445493468024,
      3.663096576477, 3.766481150018, 3.271760664144, 3.368194930794, 3.540250609025,
      3.824911768840, 3.397511985055, 3.568518356734, 3.268466398614, 3.329351334592,
      3.421372568835, 3.333950027352, 3.407773545920, 3.372240447717, 3.395095348797,
      3.375472982012, 3.407830928930, 3.389673049936, 3.444734322778, 3.434450634681,
      3.567420410364, 3.519403897623, 3.521500360692, 3.737204637428, 3.665836385359,
      3.638475909577, 3.634697522671, 3.378728206062, 3.400511009364, 3.387155190391,
      3.430448505960, 3.414568660060, 3.502974792852, 3.485472605945, 3.646792406411,
      3.600687428544, 3.596042090542, 4.043143323646, 4.042627673887, 4.000207071572,
      3.943830109299, 3.901090644538, 3.971528385050, 3.922106005834, 3.832443093990,
      3.781559803821, 3.745079273102, 3.726956468823, 4.071479016092, 4.039193036616,
      3.982347524785, 3.939905876613, 3.970318175141, 3.926366695705, 3.883604616653,
      3.829224917150, 3.786789784941, 3.137381224862, 3.001405458723, 3.186612154619,
      3.000297605619, 3.091578048072, 3.363118040231, 3.133206449617, 3.182855203960,
      3.381998093999, 3.087995448210, 2.999536488978, 3.192470183633, 3.085172887610,
      3.190455940555, 3.445493468024, 3.663096576477, 3.766481150018, 3.271760664144,
      3.368194930794, 3.540250609025, 3.824911768840, 3.397511985055, 3.568518356734,
      3.268466398614, 3.329351334592, 3.421372568835, 3.333950027352, 3.407773545920,
      3.372240447717, 3.395095348797, 3.375472982012, 3.407830928930, 3.389673049936,
      3.444734322778, 3.434450634681, 3.567420410364, 3.519403897623, 3.521500360692,
      3.737204637428, 3.665836385359, 3.638475909577, 3.634697522671, 3.378728206062,
      3.400511009364, 3.387155190391, 3.430448505960, 3.414568660060, 3.502974792852,
      3.485472605945, 3.646792406411, 3.600687428544, 3.596042090542, 4.043143323646,
      4.042627673887, 4.000207071572, 3.943830109299, 3.901090644538, 3.971528385050,
      3.922106005834, 3.832443093990, 3.781559803821, 3.745079273102, 3.726956468823,
      4.071479016092, 4.039193036616, 3.982347524785, 3.939905876613, 3.970318175141,
      3.926366695705, 3.883604616653, 3.829224917150, 3.786789784941, 3.137381224862,
      3.001405458723, 3.186612154619, 3.000297605619, 3.091578048072, 3.363118040231,
      3.133206449617, 3.182855203960, 3.381998093999, 3.087995448210, 2.999536488978,
      3.192470183633, 3.085172887610, 3.190455940555, 3.445493468024, 3.663096576477,
      3.766481150018, 3.271760664144, 3.368194930794, 3.540250609025, 3.824911768840,
      3.397511985055, 3.568518356734, 3.268466398614, 3.329351334592, 3.421372568835,
      3.333950027352, 3.407773545920, 3.372240447717, 3.395095348797, 3.375472982012,
      3.407830928930, 3.389673049936, 3.444734322778, 3.434450634681, 3.567420410364,
      3.519403897623, 3.521500360692, 3.737204637428, 3.665836385359, 3.638475909577,
      3.634697522671, 3.378728206062, 3.400511009364, 3.387155190391, 3.430448505960,
      3.414568660060, 3.502974792852, 3.485472605945, 3.646792406411, 3.600687428544,
      3.596042090542, 4.043143323646, 4.042627673887, 4.000207071572, 3.943830109299,
      3.901090644538, 3.971528385050, 3.922106005834, 3.832443093990, 3.781559803821,
      3.745079273102, 3.726956468823, 4.071479016092, 4.039193036616, 3.982347524785,
      3.939905876613, 3.970318175141, 3.926366695705, 3.883604616653, 3.829224917150,
      3.786789784941, 3.137381224862, 3.001405458723, 3.186612154619, 3.000297605619,
      3.091578048072, 3.363118040231, 3.133206449617, 3.182855203960, 3.381998093999,
      3.087995448210, 2.999536488978, 3.192470183633, 3.085172887610, 3.190455940555,
      3.445493468024, 3.663096576477, 3.766481150018, 3.271760664144, 3.368194930794,
      3.540250609025, 3.824911768840, 3.397511985055, 3.568518356734, 3.268466398614,
      3.329351334592, 3.421372568835, 3.333950027352, 3.407773545920, 3.372240447717,
      3.395095348797, 3.375472982012, 3.407830928930, 3.389673049936, 3.444734322778,
      3.434450634681, 3.567420410364, 3.519403897623, 3.521500360692, 3.737204637428,
      3.665836385359, 3.638475909577, 3.634697522671, 3.378728206062, 3.400511009364,
      3.387155190391, 3.430448505960, 3.414568660060, 3.502974792852, 3.485472605945,
      3.646792406411, 3.600687428544, 3.596042090542, 4.043143323646, 4.042627673887,
      4.000207071572, 3.943830109299, 3.901090644538, 3.971528385050, 3.922106005834,
      3.832443093990, 3.781559803821, 3.745079273102, 3.726956468823, 4.071479016092,
      4.039193036616, 3.982347524785, 3.939905876613, 3.970318175141, 3.926366695705,
      3.883604616653, 3.829224917150, 3.786789784941, 3.137381224862, 3.001405458723,
      3.186612154619, 3.000297605619, 3.091578048072, 3.363118040231, 3.133206449617,
      3.182855203960, 3.381998093999, 3.087995448210, 2.999536488978, 3.192470183633,
      3.085172887610, 3.190455940555, 3.445493468024, 3.663096576477, 3.766481150018,
      3.271760664144, 3.368194930794, 3.540250609025, 3.824911768840, 3.397511985055,
      3.568518356734, 3.268466398614, 3.329351334592, 3.421372568835, 3.333950027352,
      3.407773545920};

  return mep[i];
}

double asc_Ag_reference(int i) {
  static double asc_Ag[576] = {
      -0.002085687695, -0.002038528244, -0.003385497599, -0.002692240783,
      -0.004538826296, -0.006256935957, -0.004427462623, -0.003425534363,
      -0.005219445681, -0.003263010379, -0.003395946702, -0.005008248255,
      -0.003558712718, -0.002122854689, -0.002963787793, -0.000574041191,
      -0.004442973688, -0.001758110073, -0.005103031982, -0.006578238930,
      -0.004371465134, -0.005630175684, -0.004743123268, -0.003036162925,
      -0.004710940904, -0.003766809935, -0.002673280233, -0.003419105699,
      -0.003972574975, -0.000139297682, -0.001733573296, -0.006375867893,
      -0.005858755974, -0.003944967467, -0.002302196052, -0.001738802028,
      -0.001256552044, -0.001186089073, -0.001509544849, -0.002045416728,
      -0.006489467452, -0.007469535044, -0.005709355473, -0.003486000220,
      -0.001866862181, -0.002888446030, -0.003571046727, -0.002736100446,
      -0.004427310453, -0.004319609953, -0.001591471378, -0.003166193922,
      -0.004049882627, -0.002856209044, -0.003491796907, -0.004382752224,
      -0.002529931148, -0.004101912641, -0.003418039787, -0.004713178801,
      -0.002553583687, -0.001762365922, -0.003467423195, -0.003775482249,
      -0.002733943137, -0.003213103669, -0.005602431233, -0.002231146168,
      -0.003192121862, -0.000865638847, -0.002106560786, -0.000700644487,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000};

  return asc_Ag[i];
}

double asc_B3g_reference(int i) {
  static double asc_B3g[576] = {
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      -0.050532399154, -0.009551973174, -0.071962042607, -0.010643147094,
      -0.089783373467, -0.026341776966, -0.091266072957, -0.010662710196,
      -0.027356074353, -0.075647252100, -0.012239510683, -0.023321910317,
      -0.023579909169, -0.055103832662, -0.029685704578, -0.002608543194,
      -0.041219826083, -0.007025525578, -0.047994090188, -0.025677753050,
      -0.044269146440, -0.020378168898, -0.024350440826, -0.034219175277,
      -0.079259123585, -0.132830670580, -0.211794178476, -0.130275453807,
      -0.092727593074, -0.000900384356, -0.010616419191, -0.040408973071,
      -0.041902118584, -0.031068183493, -0.056669476873, -0.042970206780,
      -0.032940883479, -0.032495171991, -0.042068531698, -0.010473528927,
      -0.042347806882, -0.046158437117, -0.046013993667, -0.040859782010,
      -0.008632259612, -0.011817693694, -0.015399367973, -0.011144478844,
      -0.017582319314, -0.014744833006, -0.007394676012, -0.013954217983,
      -0.015037574239, -0.012429516061, -0.013965993779, -0.017524617968,
      -0.010987983891, -0.016694727259, -0.010877261182, -0.032651752502,
      -0.012526764138, -0.009237600215, -0.017692253347, -0.013515632316,
      -0.017770630907, -0.018351242050, -0.025639062891, -0.010865330530,
      -0.013561584119, -0.002834710324, -0.009548932251, -0.002495370219,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000};
  return asc_B3g[i];
}

double asc_neq_B3g_reference(int i) {
  static double asc_neq_B3g[576] = {
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      -0.050532399154, -0.009551973174, -0.071962042607, -0.010643147094,
      -0.089783373467, -0.026341776966, -0.091266072957, -0.010662710196,
      -0.027356074353, -0.075647252100, -0.012239510683, -0.023321910317,
      -0.023579909169, -0.055103832662, -0.029685704578, -0.002608543194,
      -0.041219826083, -0.007025525578, -0.047994090188, -0.025677753050,
      -0.044269146440, -0.020378168898, -0.024350440826, -0.034219175277,
      -0.079259123585, -0.132830670580, -0.211794178476, -0.130275453807,
      -0.092727593074, -0.000900384356, -0.010616419191, -0.040408973071,
      -0.041902118584, -0.031068183493, -0.056669476873, -0.042970206780,
      -0.032940883479, -0.032495171991, -0.042068531698, -0.010473528927,
      -0.042347806882, -0.046158437117, -0.046013993667, -0.040859782010,
      -0.008632259612, -0.011817693694, -0.015399367973, -0.011144478844,
      -0.017582319314, -0.014744833006, -0.007394676012, -0.013954217983,
      -0.015037574239, -0.012429516061, -0.013965993779, -0.017524617968,
      -0.010987983891, -0.016694727259, -0.010877261182, -0.032651752502,
      -0.012526764138, -0.009237600215, -0.017692253347, -0.013515632316,
      -0.017770630907, -0.018351242050, -0.025639062891, -0.010865330530,
      -0.013561584119, -0.002834710324, -0.009548932251, -0.002495370219,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000,
      0.000000000000,  0.000000000000,  0.000000000000,  0.000000000000};
  return asc_neq_B3g[i];
}

double areas_reference(int i) {
  static double areas[576] = {
      0.29141756525608, 0.19706001642314, 0.47453252418489, 0.33536257301071,
      0.65240709745199, 0.85332167762776, 0.65240709745198, 0.33864910067596,
      0.78612756823251, 0.47453252418489, 0.29354184222437, 0.65240709745198,
      0.47453252418489, 0.29141756525610, 0.41419443269637, 0.02492693051979,
      0.65240709745199, 0.15340546917602, 0.78612756823252, 0.89015303558369,
      0.65240709745199, 0.63745007977808, 0.65240709745198, 0.41419443269637,
      0.33487250307045, 0.33443860638568, 0.30059344327560, 0.33443860638523,
      0.33487250307008, 0.00100610213664, 0.09900417101450, 0.63899370140576,
      0.65240709745198, 0.47453252418489, 0.29141756525609, 0.07631247831154,
      0.03463007235347, 0.03463007235349, 0.07631247831155, 0.14972027557653,
      0.62199157092647, 0.78612756823251, 0.65240709745198, 0.41419443269637,
      0.38991489564699, 0.73732671048148, 0.73732671048148, 0.73732671048148,
      1.14932987987240, 0.62248237067432, 0.38991489564699, 0.73732671048148,
      0.73485014156995, 0.65191041220743, 0.97763931257568, 0.97763931257568,
      0.65191041220743, 0.97763931257568, 0.44301561456118, 0.71570495484030,
      0.23833601224096, 0.38991489564699, 0.73732671048148, 0.48643805106248,
      0.27736349984419, 0.65191041220743, 0.85182245635344, 0.37630532288257,
      0.50466545604027, 0.05770407057834, 0.28245796251270, 0.05369601313806,
      0.29141756525608, 0.19706001642314, 0.47453252418489, 0.33536257301071,
      0.65240709745199, 0.85332167762776, 0.65240709745198, 0.33864910067596,
      0.78612756823251, 0.47453252418489, 0.29354184222437, 0.65240709745198,
      0.47453252418489, 0.29141756525610, 0.41419443269637, 0.02492693051979,
      0.65240709745199, 0.15340546917602, 0.78612756823252, 0.89015303558369,
      0.65240709745199, 0.63745007977808, 0.65240709745198, 0.41419443269637,
      0.33487250307045, 0.33443860638568, 0.30059344327560, 0.33443860638523,
      0.33487250307008, 0.00100610213664, 0.09900417101450, 0.63899370140576,
      0.65240709745198, 0.47453252418489, 0.29141756525609, 0.07631247831154,
      0.03463007235347, 0.03463007235349, 0.07631247831155, 0.14972027557653,
      0.62199157092647, 0.78612756823251, 0.65240709745198, 0.41419443269637,
      0.38991489564699, 0.73732671048148, 0.73732671048148, 0.73732671048148,
      1.14932987987240, 0.62248237067432, 0.38991489564699, 0.73732671048148,
      0.73485014156995, 0.65191041220743, 0.97763931257568, 0.97763931257568,
      0.65191041220743, 0.97763931257568, 0.44301561456118, 0.71570495484030,
      0.23833601224096, 0.38991489564699, 0.73732671048148, 0.48643805106248,
      0.27736349984419, 0.65191041220743, 0.85182245635344, 0.37630532288257,
      0.50466545604027, 0.05770407057834, 0.28245796251270, 0.05369601313806,
      0.29141756525608, 0.19706001642314, 0.47453252418489, 0.33536257301071,
      0.65240709745199, 0.85332167762776, 0.65240709745198, 0.33864910067596,
      0.78612756823251, 0.47453252418489, 0.29354184222437, 0.65240709745198,
      0.47453252418489, 0.29141756525610, 0.41419443269637, 0.02492693051979,
      0.65240709745199, 0.15340546917602, 0.78612756823252, 0.89015303558369,
      0.65240709745199, 0.63745007977808, 0.65240709745198, 0.41419443269637,
      0.33487250307045, 0.33443860638568, 0.30059344327560, 0.33443860638523,
      0.33487250307008, 0.00100610213664, 0.09900417101450, 0.63899370140576,
      0.65240709745198, 0.47453252418489, 0.29141756525609, 0.07631247831154,
      0.03463007235347, 0.03463007235349, 0.07631247831155, 0.14972027557653,
      0.62199157092647, 0.78612756823251, 0.65240709745198, 0.41419443269637,
      0.38991489564699, 0.73732671048148, 0.73732671048148, 0.73732671048148,
      1.14932987987240, 0.62248237067432, 0.38991489564699, 0.73732671048148,
      0.73485014156995, 0.65191041220743, 0.97763931257568, 0.97763931257568,
      0.65191041220743, 0.97763931257568, 0.44301561456118, 0.71570495484030,
      0.23833601224096, 0.38991489564699, 0.73732671048148, 0.48643805106248,
      0.27736349984419, 0.65191041220743, 0.85182245635344, 0.37630532288257,
      0.50466545604027, 0.05770407057834, 0.28245796251270, 0.05369601313806,
      0.29141756525608, 0.19706001642314, 0.47453252418489, 0.33536257301071,
      0.65240709745199, 0.85332167762776, 0.65240709745198, 0.33864910067596,
      0.78612756823251, 0.47453252418489, 0.29354184222437, 0.65240709745198,
      0.47453252418489, 0.29141756525610, 0.41419443269637, 0.02492693051979,
      0.65240709745199, 0.15340546917602, 0.78612756823252, 0.89015303558369,
      0.65240709745199, 0.63745007977808, 0.65240709745198, 0.41419443269637,
      0.33487250307045, 0.33443860638568, 0.30059344327560, 0.33443860638523,
      0.33487250307008, 0.00100610213664, 0.09900417101450, 0.63899370140576,
      0.65240709745198, 0.47453252418489, 0.29141756525609, 0.07631247831154,
      0.03463007235347, 0.03463007235349, 0.07631247831155, 0.14972027557653,
      0.62199157092647, 0.78612756823251, 0.65240709745198, 0.41419443269637,
      0.38991489564699, 0.73732671048148, 0.73732671048148, 0.73732671048148,
      1.14932987987240, 0.62248237067432, 0.38991489564699, 0.73732671048148,
      0.73485014156995, 0.65191041220743, 0.97763931257568, 0.97763931257568,
      0.65191041220743, 0.97763931257568, 0.44301561456118, 0.71570495484030,
      0.23833601224096, 0.38991489564699, 0.73732671048148, 0.48643805106248,
      0.27736349984419, 0.65191041220743, 0.85182245635344, 0.37630532288257,
      0.50466545604027, 0.05770407057834, 0.28245796251270, 0.05369601313806,
      0.29141756525608, 0.19706001642314, 0.47453252418489, 0.33536257301071,
      0.65240709745199, 0.85332167762776, 0.65240709745198, 0.33864910067596,
      0.78612756823251, 0.47453252418489, 0.29354184222437, 0.65240709745198,
      0.47453252418489, 0.29141756525610, 0.41419443269637, 0.02492693051979,
      0.65240709745199, 0.15340546917602, 0.78612756823252, 0.89015303558369,
      0.65240709745199, 0.63745007977808, 0.65240709745198, 0.41419443269637,
      0.33487250307045, 0.33443860638568, 0.30059344327560, 0.33443860638523,
      0.33487250307008, 0.00100610213664, 0.09900417101450, 0.63899370140576,
      0.65240709745198, 0.47453252418489, 0.29141756525609, 0.07631247831154,
      0.03463007235347, 0.03463007235349, 0.07631247831155, 0.14972027557653,
      0.62199157092647, 0.78612756823251, 0.65240709745198, 0.41419443269637,
      0.38991489564699, 0.73732671048148, 0.73732671048148, 0.73732671048148,
      1.14932987987240, 0.62248237067432, 0.38991489564699, 0.73732671048148,
      0.73485014156995, 0.65191041220743, 0.97763931257568, 0.97763931257568,
      0.65191041220743, 0.97763931257568, 0.44301561456118, 0.71570495484030,
      0.23833601224096, 0.38991489564699, 0.73732671048148, 0.48643805106248,
      0.27736349984419, 0.65191041220743, 0.85182245635344, 0.37630532288257,
      0.50466545604027, 0.05770407057834, 0.28245796251270, 0.05369601313806,
      0.29141756525608, 0.19706001642314, 0.47453252418489, 0.33536257301071,
      0.65240709745199, 0.85332167762776, 0.65240709745198, 0.33864910067596,
      0.78612756823251, 0.47453252418489, 0.29354184222437, 0.65240709745198,
      0.47453252418489, 0.29141756525610, 0.41419443269637, 0.02492693051979,
      0.65240709745199, 0.15340546917602, 0.78612756823252, 0.89015303558369,
      0.65240709745199, 0.63745007977808, 0.65240709745198, 0.41419443269637,
      0.33487250307045, 0.33443860638568, 0.30059344327560, 0.33443860638523,
      0.33487250307008, 0.00100610213664, 0.09900417101450, 0.63899370140576,
      0.65240709745198, 0.47453252418489, 0.29141756525609, 0.07631247831154,
      0.03463007235347, 0.03463007235349, 0.07631247831155, 0.14972027557653,
      0.62199157092647, 0.78612756823251, 0.65240709745198, 0.41419443269637,
      0.38991489564699, 0.73732671048148, 0.73732671048148, 0.73732671048148,
      1.14932987987240, 0.62248237067432, 0.38991489564699, 0.73732671048148,
      0.73485014156995, 0.65191041220743, 0.97763931257568, 0.97763931257568,
      0.65191041220743, 0.97763931257568, 0.44301561456118, 0.71570495484030,
      0.23833601224096, 0.38991489564699, 0.73732671048148, 0.48643805106248,
      0.27736349984419, 0.65191041220743, 0.85182245635344, 0.37630532288257,
      0.50466545604027, 0.05770407057834, 0.28245796251270, 0.05369601313806,
      0.29141756525608, 0.19706001642314, 0.47453252418489, 0.33536257301071,
      0.65240709745199, 0.85332167762776, 0.65240709745198, 0.33864910067596,
      0.78612756823251, 0.47453252418489, 0.29354184222437, 0.65240709745198,
      0.47453252418489, 0.29141756525610, 0.41419443269637, 0.02492693051979,
      0.65240709745199, 0.15340546917602, 0.78612756823252, 0.89015303558369,
      0.65240709745199, 0.63745007977808, 0.65240709745198, 0.41419443269637,
      0.33487250307045, 0.33443860638568, 0.30059344327560, 0.33443860638523,
      0.33487250307008, 0.00100610213664, 0.09900417101450, 0.63899370140576,
      0.65240709745198, 0.47453252418489, 0.29141756525609, 0.07631247831154,
      0.03463007235347, 0.03463007235349, 0.07631247831155, 0.14972027557653,
      0.62199157092647, 0.78612756823251, 0.65240709745198, 0.41419443269637,
      0.38991489564699, 0.73732671048148, 0.73732671048148, 0.73732671048148,
      1.14932987987240, 0.62248237067432, 0.38991489564699, 0.73732671048148,
      0.73485014156995, 0.65191041220743, 0.97763931257568, 0.97763931257568,
      0.65191041220743, 0.97763931257568, 0.44301561456118, 0.71570495484030,
      0.23833601224096, 0.38991489564699, 0.73732671048148, 0.48643805106248,
      0.27736349984419, 0.65191041220743, 0.85182245635344, 0.37630532288257,
      0.50466545604027, 0.05770407057834, 0.28245796251270, 0.05369601313806,
      0.29141756525608, 0.19706001642314, 0.47453252418489, 0.33536257301071,
      0.65240709745199, 0.85332167762776, 0.65240709745198, 0.33864910067596,
      0.78612756823251, 0.47453252418489, 0.29354184222437, 0.65240709745198,
      0.47453252418489, 0.29141756525610, 0.41419443269637, 0.02492693051979,
      0.65240709745199, 0.15340546917602, 0.78612756823252, 0.89015303558369,
      0.65240709745199, 0.63745007977808, 0.65240709745198, 0.41419443269637,
      0.33487250307045, 0.33443860638568, 0.30059344327560, 0.33443860638523,
      0.33487250307008, 0.00100610213664, 0.09900417101450, 0.63899370140576,
      0.65240709745198, 0.47453252418489, 0.29141756525609, 0.07631247831154,
      0.03463007235347, 0.03463007235349, 0.07631247831155, 0.14972027557653,
      0.62199157092647, 0.78612756823251, 0.65240709745198, 0.41419443269637,
      0.38991489564699, 0.73732671048148, 0.73732671048148, 0.73732671048148,
      1.14932987987240, 0.62248237067432, 0.38991489564699, 0.73732671048148,
      0.73485014156995, 0.65191041220743, 0.97763931257568, 0.97763931257568,
      0.65191041220743, 0.97763931257568, 0.44301561456118, 0.71570495484030,
      0.23833601224096, 0.38991489564699, 0.73732671048148, 0.48643805106248,
      0.27736349984419, 0.65191041220743, 0.85182245635344, 0.37630532288257,
      0.50466545604027, 0.05770407057834, 0.28245796251270, 0.05369601313806};
  return areas[i];
}

#endif /* C_HOST_FUNCTIONS_H */
